/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
//


#include "ns3/log.h"
#include "ns3/address.h"
#include "ns3/node.h"
#include "ns3/nstime.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/tag.h"
#include "ns3/uinteger.h"
#include "ns3/integer.h"
#include "ns3/string.h"
#include "ns3/boolean.h"
#include "ns3/trace-source-accessor.h"
#include "traffic-light.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/inet-socket-address.h"
#include "ns3/double.h"
#include "ns3/node-list.h"

#include <math.h>
#include "ns3/virtualsource11p-propagation-loss-model.h"

NS_LOG_COMPONENT_DEFINE ("TrafficLight");

using namespace std;

namespace ns3 {


NS_OBJECT_ENSURE_REGISTERED (TrafficLight);

TypeId
TrafficLight::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TrafficLight")
    .SetParent<Application> ()
    .AddConstructor<TrafficLight> ()
    .AddAttribute ("Interval",
                   "The time interval between two successive broadcast transmissions.",
                   TimeValue (Seconds(0.1)),
                   MakeTimeAccessor (&TrafficLight::m_interval),
                   MakeTimeChecker())
    .AddAttribute ("PhaseLength",
                   "Minimum length of traffic light phase",
                   TimeValue (Seconds(30)),
                   MakeTimeAccessor (&TrafficLight::m_phaseLength),
                   MakeTimeChecker())
    .AddAttribute ("RemoteAddress",
                   "The address of the destination (either a Unicast or a Broadcast address)",
                   AddressValue (),
                   MakeAddressAccessor (&TrafficLight::m_peer),
                   MakeAddressChecker ())
    .AddAttribute ("Protocol",
                   "The type of protocol to use.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&TrafficLight::m_protocol),
                   MakeTypeIdChecker ())
    .AddAttribute ("Synchronized",
                   "All traffic lights are synchronized.",
                   BooleanValue (true),
                   MakeBooleanAccessor (&TrafficLight::m_sync),
                   MakeBooleanChecker ())
    ;
  return tid;
}


TrafficLight::TrafficLight()
 : m_socket(0),
   m_seqNo(0),
   m_id (0)
{
  NS_LOG_FUNCTION_NOARGS ();
  
  NS_ASSERT (IDM_INT_INVALID == VTL_INT_INVALID);
  
}

TrafficLight::~TrafficLight()
{
  NS_LOG_FUNCTION_NOARGS ();
}

void
TrafficLight::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_socket = 0;

    // chain up
  Application::DoDispose ();

}




void
TrafficLight::StartApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  
  // init position and direction
  m_id = VTL_INT_INVALID; 
  
  
    // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode(), m_protocol);
      m_socket->Bind ();
      m_socket->Connect (m_peer);
      
    }
    
    // Ensure no pending event
  CancelEvents ();

  m_intersectionId = m_int->m_id;
  GetNode()->GetObject<MobilityModel>()->SetPosition(m_int->m_pos);
  NS_LOG_DEBUG("Starting TrafficLight at intersection " << m_intersectionId 
  << " at " << GetNode()->GetObject<MobilityModel>()->GetPosition());

  InitVtl();
  
    // Start application with jitter
  UniformVariable jitterGenerator(0, m_interval.GetSeconds ());
  double jitterInSeconds = jitterGenerator.GetValue();
  Time m_startJitter = Seconds(jitterInSeconds);

  m_sendEvent = Simulator::Schedule(m_startJitter, &TrafficLight::NextTimestep, this);
}

void
TrafficLight::StopApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
  CancelEvents ();
  if(m_socket != 0)
    {
      m_socket->Close ();
    }
  else
    {
      NS_LOG_WARN("TrafficLight found null socket to close in StopApplication");
    }
}

void
TrafficLight::CancelEvents ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Simulator::Cancel(m_sendEvent);
  Simulator::Cancel(m_leaderUpdateEvent);
}

void
TrafficLight::ScheduleNextTx()
{
  NS_LOG_FUNCTION_NOARGS ();
  Time nextTime(m_interval);
  m_sendEvent = Simulator::Schedule(nextTime, &TrafficLight::NextTimestep, this);
}



void
TrafficLight::VtlExpire (void)
{
  if (m_vtlList.empty ())
    return;

    
  if (m_vtlList.front().validUntil <= Simulator::Now ()) {
    m_vtlList.pop_front ();
  }
}




void
TrafficLight::InitVtl ()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  m_vtlList.clear();
  m_roadmap = m_int->GetRoadMap ();

  std::vector<std::list<Ptr<Road> > >::iterator it;
  std::list<Ptr<Road> >::iterator rit;
  NS_ASSERT(m_roadmap.size () > 0);

  if (m_sync) {
    m_currentPhase = 0;
    LeaderUpdateTimeslice ();     
  } else {
    UniformVariable phaserandom = UniformVariable(0, m_roadmap.size ());
    m_currentPhase = (int)phaserandom.GetValue();
    
    UniformVariable jitterGenerator = UniformVariable(0, m_phaseLength.GetSeconds ());
    double jitterInSeconds = jitterGenerator.GetValue();
    Simulator::Schedule(Seconds(jitterInSeconds), &TrafficLight::LeaderUpdateTimeslice, this);
  }
}

void
TrafficLight::LeaderUpdateTimeslice()
{
  NS_ASSERT (m_leaderUpdateEvent.IsExpired ());

    
  Time validFrom;
  std::list<Ptr<Road> >::iterator rit;
  
  m_currentPhase++;
  if (m_currentPhase == m_roadmap.size ())
    m_currentPhase = 0;
  
  
  if (m_vtlList.empty()) {
    validFrom = Simulator::Now ();
  } else {
    validFrom = m_vtlList.back().validUntil;
  }
  VtlInfo vi;
  
  vi.validFrom = validFrom;
  vi.validUntil = validFrom + m_phaseLength;
  memset(&vi.lightmap, 0, sizeof(vi.lightmap));
  
  for (rit = m_roadmap[m_currentPhase].begin(); rit != m_roadmap[m_currentPhase].end(); ++rit) {
    vi.lightmap[(*rit)->m_localId] = 1;
  }
  m_vtlList.push_back(vi);
  Simulator::Schedule(vi.validUntil - Simulator::Now (), &TrafficLight::VtlExpire, this);
  
  
  
  m_leaderUpdateEvent = Simulator::Schedule(vi.validUntil - Seconds(5) - Simulator::Now (), &TrafficLight::LeaderUpdateTimeslice, this);
}





void 
TrafficLight::NextTimestep ()
{
  NS_LOG_FUNCTION_NOARGS ();
 
  SendPacket();
  ScheduleNextTx();


}




void
TrafficLight::SendPacket()
{
  NS_LOG_FUNCTION_NOARGS ();
//  NS_LOG_INFO ("VtlApplication::SendPacket() at " << Simulator::Now().GetSeconds() << "sec");
  NS_ASSERT (m_sendEvent.IsExpired ());
  Ptr<Packet> packet;

  packet = Create<Packet> ();;

  
    // Attach VTL header to packet
  VtlHeader vtlHdr;
  vtlHdr.m_source = m_id;
  vtlHdr.m_seqNo = m_seqNo++;
  vtlHdr.m_timestamp = Simulator::Now();
  vtlHdr.m_intersectionId = m_intersectionId;
  vtlHdr.m_localId = VTL_BYTE_INVALID;
  
  vtlHdr.m_nextLeader = VTL_INT_INVALID;



  uint32_t i = 0;
  std::list<VtlInfo>::iterator it;
  for (it = m_vtlList.begin(); it != m_vtlList.end(); ++it) {
    vtlHdr.m_vtlInfo[i++] = *it;
  }
  vtlHdr.m_nVtlInfo = i;
  
  
  packet->AddHeader(vtlHdr);


  int rc;
  rc = m_socket->Send (packet);
  
  if (rc < 0) {
    NS_LOG_DEBUG ( "###################socket error" << (Socket::SocketErrno) m_socket->GetErrno());
  }
  

}


} // Namespace ns3
