/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 *
 */
#ifndef MANHATTAN_IDM_H
#define MANHATTAN_IDM_H

#include "mobility-model.h"
#include "ns3/nstime.h"
#include "ns3/node.h"
#include "ns3/event-id.h"
#include "ns3/vector.h"
#include <map>
#include <queue>
#include <list>
#include <vector>


#define IDM_INVALID_POS 999999999
#define IDM_INT_INVALID 999999999
#define IDM_BYTE_INVALID 255
#define IDM_HIGH_PRIO 255
#define IDM_MAX_ROADS_PER_INTERSECTION 4



namespace ns3 {

class IDMUtils
{
public:
 static ns3::Vector vplus(const Vector &v, const Vector &w);
 static ns3::Vector vmult(const Vector &v, double d);
 static ns3::Vector vdiff(const Vector &v, const Vector &w);
 static ns3::Vector vnorm(const Vector &v);
 static double vlength(const Vector &v);
 static double vdist (const Vector &a, const Vector &b);
};

class IntelligentDriverMobilityModel;
class Road;


/**
 * \ingroup mobility
 *
 * \brief Intersection Model
 */
class Intersection : public Object 
{
public:
  struct queueEntry {
    uint32_t  vid;
    uint32_t  rid;
  };
  
  static TypeId GetTypeId (void);

  Intersection ();
  ~Intersection ();

  /* road conflict handling  */
  void                      AddNoConflict (Ptr<Road>, Ptr<Road>);
  void                      AddNoConflict (void);
  bool                      IsRoadConflicting (uint32_t, uint32_t);
  std::vector<std::list<Ptr<Road> > >     GetRoadMap (void);
  void                      MakeRoadMap (void);
  
  
  /* priority handling */
  bool        CanGo (Ptr<Road>, uint32_t);
  void        AddToQueue (Ptr<Road>, uint32_t);
  void        AddToQueueFront (Ptr<Road>, uint32_t);
  bool        RemoveQueued (Ptr<Road> road, uint32_t);
  void        PrintQueues (void);
  
  /* set helper */
  void        AddApproachRoad (Ptr<Road>);
  static bool        CompareRoadPrio (Ptr<Road> r1, Ptr<Road> r2);
  Ptr<Road>   GetRoadByLocalId(uint8_t lid);
  
  
  uint32_t    m_id;
  Vector      m_pos;
  double      m_ibd;
  bool        m_suburban;
  double      m_stopDist;
  uint8_t     m_localIdCount;
  Ptr<Road>   m_lastAdded;
  Ptr<Road>   m_lastAdded2;
  
  std::list<Ptr<Road> > m_roadsTo;
  std::list<Ptr<Road> > m_roadsFrom;
  
  std::map<uint8_t,std::list<queueEntry> > m_queues;
  std::map<uint32_t,std::list<Ptr<Road> > > m_noconflict;
  
  Ptr<Road>   m_localIdToRoad[IDM_MAX_ROADS_PER_INTERSECTION];
  
  // 2 dimensional list, each entry contains all non-conflicting roads
  // sorted descending from highest prio to lowest
  // can be used for traffic light assignment
  std::vector<std::list<Ptr<Road> > > m_rl;
  
  
}; // class Intersection


/**
 * \ingroup mobility
 *
 * \brief Road
 */
class Navigator;
 
class Road : public Object 
{
public:

  enum roadrelation {
    LOS,
    RIGHT,
    LEFT
  };
  
  
  static TypeId GetTypeId (void);
  static std::vector<Ptr<Road> > roads;

  Road ();
  ~Road ();
  
  void          addVehicle(Ptr<Navigator>);
  double        GetGapTime();
  void          CheckNextIntersection ();
  double        GetHighRoadGap ();
  Ptr<Road>     GetHighPrioRoad (void);
  roadrelation  GetRelationTo (Ptr<Road> other);
  
  uint32_t    m_id;
  uint8_t     m_localId;  // id at crossing m_end
  Vector      m_startPos;
  double      m_length;
  Vector      m_dir;
  Ptr<Intersection>   m_start;
  Ptr<Intersection>   m_end;
  uint8_t             m_prio;
  
  // list of vehicles currently on the road
  std::queue<Ptr<Navigator> > m_vehicles;

  // other roads to end intersection with higher prio than this road
  std::vector<Ptr<Road> > m_roadsHiPrio;  
  

  
  
}; // class Road



class Navigator : public Object 
{
public:
  static TypeId GetTypeId (void);

  Navigator (Ptr<Road> source);
  Navigator ();
  ~Navigator ();
  
  enum RoadState {
    UNCHANGED,
    CHANGED,
    FINISHED
  };

  enum FailsafeState {
    INIT,
    APPROACH,
    WAITING,
    CROSSING,
    CROSSING2
  };

  void                                  Start();
  RoadState                             CheckRoadChange (double);
  Vector                                GetCurrentDirection ();
  Vector                                GetCurrentRoadStart ();
  Ptr<IntelligentDriverMobilityModel>   GetCurrentPOV ();
  double                                GetStopDist ();
  double                                GetCompleteCrossingDist ();
  bool                                  IsFirstVehicle ();
  double                                GetTimeToIntersection ();
  bool                                  CheckGo (double requiredGap);
  void                                  checkFailsafe (double speed, double requiredGap);
  void                                  EndFailsafe ();
  void                                  UpdateRoad (Ptr<Road> r);
  Ptr<Intersection>                     GetCommonIntersection (Ptr<Navigator>);
  uint32_t                              GetCurrentIntersectionId (void);
  uint8_t                               GetCurrentLocalId (void);
  void                                  SetUpdateCallback (Callback<void, uint32_t, uint8_t>);
  Road::roadrelation                    GetRelationTo (uint8_t otherid);
  
  
  Ptr<Road>                             m_curRoad;
  Ptr<Intersection>                     m_curInt;
  Ptr<Road>                             m_lastRoad;
  Ptr<Intersection>                     m_lastInt;
  Ptr<IntelligentDriverMobilityModel>   m_idm;
  Ptr<Navigator>                        m_pov;
  uint32_t                              m_id;
  FailsafeState                         m_astate;
  bool                                  m_running;
private:
  double                                m_carLength;
  
  Callback<void, uint32_t, uint8_t>     m_roadUpdate;
  
}; // class Navigator



/**
 * \ingroup mobility
 *
 * \brief Intelligent driver car following model.
 */
class IntelligentDriverMobilityModel : public MobilityModel 
{
public:

  enum ApproachState {
    FAILSAFE,
    RED,
    GREEN
  };

  static TypeId GetTypeId (void);
  
  /**
   * Create position located at coordinates (0,0,0) with
   * speed (0,0,0).
   */
  IntelligentDriverMobilityModel ();
  virtual ~IntelligentDriverMobilityModel ();
  void SetDirection (Vector &dir);
  void SetDesiredSpeed (double speed);
  void StopAt (double distance);
  void StopAt (const Vector &position);
  bool GetStop (void);
  void Resume (void);
  void EndMobility (void);
  void SetApproach (ApproachState);
  double GetSpeed (void);
  double GetAcc (void);
  double GetCarLength (void);
  Ptr<Navigator> m_navi;
  uint32_t    m_myId;
  double  m_distDriven;

private:
  virtual Vector DoGetPosition (void) const;
  virtual void DoSetPosition (const Vector &position);
  virtual Vector DoGetVelocity (void) const;
  
  void SetStartTimeEvent (void);

  void NextTimestep (void);
  void UpdatePositionAndSpeed (void);
  void GetHeadingVehicleSpeedAndDist (void);
  


  Time m_baseTime;
  Vector m_pos;
  double m_speed;
  double m_acc;
  Vector m_dir;
  Vector m_stoppos;
  bool   m_stop;
  uint32_t    m_povId;
  bool    m_end;
  double  m_accLimit;
  ApproachState    m_astate;
  
  // idm
  double m_desSpeed;
  double m_minSpace;
  double m_desHeadway;
  double m_desDec;
  double m_accExp;
  double m_carLength;
  double m_desAcc;
  Ptr<MobilityModel> m_povmob;
  
  double m_headspeed;
  double m_headdist;
  
  
  EventId       m_nextEvent;
  Time          m_interval;   // the time interval
  Time          m_startTime;
  Time          m_stopTime;
  
  
  TracedCallback<Time,uint32_t,Vector,Vector,double> m_positionTrace;
  
};




} // namespace ns3

#endif /* MANHATTAN_IDM_H */
